"""
summarizer.py

This module provides summarization functionality for indexed tables.
"""

import gc
import json
import logging
import math
from collections import defaultdict

import duckdb
import fire
import pandas as pd
import tiktoken
import torch
from openai import OpenAI
from sentence_transformers import SentenceTransformer
from tqdm import tqdm
from transformers import TextGenerationPipeline

from pneuma.utils.logging_config import configure_logging
from pneuma.utils.prompting_interface import (
    prompt_openai_llm,
    prompt_pipeline,
    prompt_pipeline_robust,
)
from pneuma.utils.response import Response, ResponseStatus
from pneuma.utils.summary_types import SummaryType
from pneuma.utils.table_status import TableStatus

configure_logging()
logger = logging.getLogger("Summarizer")


class Summarizer:
    """
    Summarizes indexed tables in the database.

    This class provides a method to summarize indexed tables in the database
    to represent them for retrieval purposes.

    ## Attributes
    - **pipe** (`OpenAI | TextGenerationPipeline`): The LLM pipeline for inference.
    - **embedding_model** (`OpenAI | SentenceTransformer`): The model used for
    text embeddings.
    - **db_path** (`str`): Path to the database file for retrieving content
    summaries & context.
    - **MAX_LLM_BATCH_SIZE** (`int`): The upper bound of batch size value to
    explore dynamically for LLM inference.
    - **EMBEDDING_MAX_TOKENS** (`int`): The maximum number of tokens the embedding
    model supports (hard-coded to 512 for local models and 8191 for OpenAI models).
    """
    def __init__(
        self,
        llm: OpenAI | TextGenerationPipeline,
        embed_model: OpenAI | SentenceTransformer,
        db_path: str,
        max_llm_batch_size: int = 50,
    ):
        self.db_path = db_path
        self.pipe = llm
        self.embedding_model = embed_model
        self.MAX_LLM_BATCH_SIZE = max_llm_batch_size

        if isinstance(self.embedding_model, OpenAI):
            self.EMBEDDING_MAX_TOKENS = 8191
        else:
            self.EMBEDDING_MAX_TOKENS = 512

    def summarize(self, table_id: str = None) -> str:
        """
        Summarizes the contents of all unsummarized tables or a specific table
        if `table_id` is provided.

        ## Args
        - **table_id** (`str`): The specific table ID to be summarized.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                if table_id is None or table_id == "":
                    logger.info("Generating summaries for all unsummarized tables")
                    table_ids = [
                        entry[0].replace("'", "''")
                        for entry in connection.sql(
                            f"""SELECT id FROM table_status
                            WHERE status = '{TableStatus.REGISTERED}'"""
                        ).fetchall()
                    ]
                    logger.info(f"Found {len(table_ids)} unsummarized tables")
                else:
                    table_ids = [table_id.replace("'", "''")]

                if len(table_ids) == 0:
                    return Response(
                        status=ResponseStatus.SUCCESS,
                        message="No unsummarized tables found.\n",
                        data={"table_ids": []},
                    ).to_json()
                if len(table_ids) == 1:
                    all_summary_ids = self.__summarize_table_by_id(table_ids[0])
                else:
                    all_summary_ids = self.__batch_summarize_tables(table_ids)

                if len(all_summary_ids) == 0:
                    return Response(
                        status=ResponseStatus.ERROR,
                        message=f"Summarization failed",
                    ).to_json()

                return Response(
                    status=ResponseStatus.SUCCESS,
                    message=f"Total of {len(all_summary_ids)} summaries has been added "
                    f"with IDs: {', '.join([str(summary_id) for summary_id in all_summary_ids])}.\n",
                    data={"table_ids": table_ids, "summary_ids": all_summary_ids},
                ).to_json()
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error connecting to database: {e}",
            ).to_json()

    def __summarize_table_by_id(self, table_id: str) -> list[str]:
        """
        Summarizes the contents of a single table: `table_id`.

        ## Args
        - **table_id** (`str`): The specific table ID to be summarized.

        ## Returns
        - `list[str]`: The database IDs of the resulting summaries for table
        `table_id`.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                status = connection.sql(
                    f"SELECT status FROM table_status WHERE id = '{table_id}'"
                ).fetchone()[0]
                if status == str(TableStatus.SUMMARIZED) or status == str(
                    TableStatus.DELETED
                ):
                    logger.warning(
                        "Table with ID %s has already been summarized.", table_id
                    )
                    return []

                table_df = connection.sql(f"SELECT * FROM '{table_id}'").to_df()

                narration_summaries = self.__generate_column_narrations(table_df)
                row_samples = self.__generate_row_samples(table_df)

                summary_ids = []

                for narration_summary in narration_summaries:
                    narration_payload = json.dumps({"payload": narration_summary})
                    narration_payload = narration_payload.replace("'", "''")
                    summary_ids.append(
                        connection.sql(
                            f"""INSERT INTO table_summaries (table_id, summary, summary_type)
                            VALUES ('{table_id}', '{narration_payload}', '{SummaryType.COLUMN_NARRATION}')
                            RETURNING id"""
                        ).fetchone()[0]
                    )

                for row_sample in row_samples:
                    row_payload = json.dumps({"payload": row_sample})
                    row_payload = row_payload.replace("'", "''")
                    summary_ids.append(
                        connection.sql(
                            f"""INSERT INTO table_summaries (table_id, summary, summary_type)
                            VALUES ('{table_id}', '{row_payload}', '{SummaryType.ROW_SAMPLE}')
                            RETURNING id"""
                        ).fetchone()[0]
                    )

                connection.sql(
                    f"""UPDATE table_status
                    SET status = '{TableStatus.SUMMARIZED}'
                    WHERE id = '{table_id}'"""
                )

                return summary_ids
        except Exception as e:
            logger.error(f"Error connecting to database: {e}")
            return []

    def __batch_summarize_tables(self, table_ids: list[str]) -> list[str]:
        """
        Summarizes the contents of tables `table_ids`.

        ## Args
        - **table_ids** (`list[str]`): The specific table IDs to be summarized.

        ## Returns
        - `list[str]`: The database IDs of the resulting summaries for the tables.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                for table_id in table_ids:
                    status = connection.sql(
                        f"SELECT status FROM table_status WHERE id = '{table_id}'"
                    ).fetchone()[0]
                    if status == str(TableStatus.SUMMARIZED) or status == str(
                        TableStatus.DELETED
                    ):
                        logger.warning(
                            "Table with ID %s has already been summarized.", table_id
                        )
                        table_ids.remove(table_id)

                all_narration_summaries = self.__batch_generate_column_narrations(
                    table_ids
                )
                summary_ids = []

                for table_id, narration_summaries in all_narration_summaries.items():
                    table_df = connection.sql(f"SELECT * FROM '{table_id}'").to_df()
                    row_samples = self.__generate_row_samples(table_df)

                    for narration_summary in narration_summaries:
                        narration_payload = json.dumps({"payload": narration_summary})
                        narration_payload = narration_payload.replace("'", "''")
                        summary_ids.append(
                            connection.sql(
                                f"""INSERT INTO table_summaries (table_id, summary, summary_type)
                                VALUES ('{table_id}', '{narration_payload}', '{SummaryType.COLUMN_NARRATION}')
                                RETURNING id"""
                            ).fetchone()[0]
                        )

                    for row_sample in row_samples:
                        row_payload = json.dumps({"payload": row_sample})
                        row_payload = row_payload.replace("'", "''")
                        summary_ids.append(
                            connection.sql(
                                f"""INSERT INTO table_summaries (table_id, summary, summary_type)
                                VALUES ('{table_id}', '{row_payload}', '{SummaryType.ROW_SAMPLE}')
                                RETURNING id"""
                            ).fetchone()[0]
                        )

                    connection.sql(
                        f"""UPDATE table_status
                        SET status = '{TableStatus.SUMMARIZED}'
                        WHERE id = '{table_id}'"""
                    )

                return summary_ids
        except Exception as e:
            logger.error(f"Error connecting to database: {e}")
            return []

    def __generate_column_narrations(self, df: pd.DataFrame) -> list[str]:
        """Generate column narrations for a single dataframe (for quick local
        testing). This method may be removed in the future."""
        cols = df.columns
        conversations = []
        for col in cols:
            prompt = self.__get_col_narration_prompt(" | ".join(cols), col)
            conversations.append([{"role": "user", "content": prompt}])

        if len(conversations) > 0:
            if isinstance(self.pipe, OpenAI):
                outputs = prompt_openai_llm(
                    llm=self.pipe,
                    conversations=conversations,
                    max_new_tokens=400,
                )
            else:
                outputs = prompt_pipeline(
                    self.pipe,
                    conversations,
                    batch_size=2,
                    context_length=32768,
                    max_new_tokens=400,
                    temperature=None,
                    top_p=None,
                    top_k=None,
                )

            col_narrations: list[str] = []
            for output_idx, output in enumerate(outputs):
                col_narrations.append(
                    f"{cols[output_idx]}: {output[-1]['content']}".strip()
                )

        merged_column_descriptions = self.__block_column_narrations(col_narrations)
        return merged_column_descriptions

    def __batch_generate_column_narrations(
        self, table_ids: list[str]
    ) -> dict[str, list[str]]:
        """
        Generates column narrations for the tables `table_ids`.

        ## Args
        - **table_ids** (`list[str]`): The specific table IDs to be narrated.

        ## Returns
        - `dict[str, list[str]]`: The column narrations of the tables.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                summaries: dict[str, list[str]] = {}
                conversations = []
                conv_tables = []
                conv_cols = []

                for table_id in table_ids:
                    table_df = connection.sql(f"SELECT * FROM '{table_id}'").to_df()
                    cols = table_df.columns
                    for col in cols:
                        prompt = self.__get_col_narration_prompt(" | ".join(cols), col)
                        conversations.append([{"role": "user", "content": prompt}])
                        conv_tables.append(table_id)
                        conv_cols.append(col)

                if isinstance(self.pipe, OpenAI):
                    optimal_batch_size = 1
                    sorted_indices = list(range(len(conversations)))
                else:
                    optimal_batch_size = self.__get_optimal_batch_size(conversations)
                    sorted_indices = self.__get_special_indices(
                        conversations, optimal_batch_size
                    )

                conversations = [conversations[i] for i in sorted_indices]
                conv_tables = [conv_tables[i] for i in sorted_indices]
                conv_cols = [conv_cols[i] for i in sorted_indices]

                if len(conversations) > 0:
                    if isinstance(self.pipe, OpenAI):
                        outputs = prompt_openai_llm(
                            llm=self.pipe,
                            conversations=conversations,
                            max_new_tokens=400,
                        )
                    else:
                        outputs: list[list[dict[str, str]]] = []
                        max_batch_size = optimal_batch_size
                        same_batch_size_counter = 0
                        for i in tqdm(range(0, len(conversations), optimal_batch_size)):
                            llm_output = prompt_pipeline_robust(
                                self.pipe,
                                conversations[i : i + optimal_batch_size],
                                batch_size=optimal_batch_size,
                                context_length=32768,
                                max_new_tokens=400,
                                temperature=None,
                                top_p=None,
                                top_k=None,
                            )
                            outputs += llm_output[0]

                            if llm_output[1] == optimal_batch_size:
                                same_batch_size_counter += 1
                                if same_batch_size_counter % 10 == 0:
                                    optimal_batch_size = min(
                                        optimal_batch_size + 2, max_batch_size
                                    )
                            else:
                                optimal_batch_size = llm_output[1]
                                same_batch_size_counter = 0

                    col_narrations: dict[str, list[str]] = defaultdict(list)
                    for output_idx, output in enumerate(outputs):
                        col_narrations[conv_tables[output_idx]] += [
                            f"{conv_cols[output_idx]}: {output[-1]['content']}".strip()
                        ]

                for key, value in col_narrations.items():
                    summaries[key] = self.__block_column_narrations(value)

                return summaries
        except Exception as e:
            logger.error(f"Error connecting to database: {e}")
            return {}

    def __get_col_narration_prompt(self, columns: str, column: str) -> str:
        """
        Returns the prompt to narrate a column of a table given other columns
        in the table.

        ## Args
        - **columns** (`str`): A concatenation of `columns`.
        - **column** (`str`): A specific column (part of `columns`) to be narrated.

        ## Returns
        - `str`: The prompt to narrate `column`.
        """
        return f"""A table has the following columns:
/*
{columns}
*/
Describe briefly what the {column} column represents. If not possible, simply state "No description.\""""

    def __get_optimal_batch_size(self, conversations: list[dict[str, str]]) -> int:
        """
        Explores the optimal batch size value (bounded between 1 and
        `MAX_LLM_BATCH_SIZE`) for `conversations` to be set for the LLM pipeline
        using binary search.

        ## Args
        - **conversations** (`list[dict[str, str]]`): The list of prompts to
        narrate columns of tables.

        ## Returns
        - `int`: The optimal batch size.
        """
        max_batch_size = self.MAX_LLM_BATCH_SIZE
        min_batch_size = 1
        logger.info(
            f"Exploring optimal batch size within [{min_batch_size},{max_batch_size}] (binary search)"
        )

        while min_batch_size < max_batch_size:
            mid_batch_size = (min_batch_size + max_batch_size) // 2
            logger.info(f"=> Current mid batch size: {mid_batch_size}")
            if self.__is_fit_in_memory(conversations, mid_batch_size):
                min_batch_size = mid_batch_size + 1
            else:
                max_batch_size = mid_batch_size - 1
        optimal_batch_size = min_batch_size
        logger.info(f"Optimal batch size: {optimal_batch_size}")
        return optimal_batch_size

    def __is_fit_in_memory(
        self, conversations: list[dict[str, str]], batch_size: int
    ) -> bool:
        """
        Checks if `conversations` with the given `batch_size` fits in memory when
        running inference using the LLM pipeline.

        ## Args
        - **conversations** (`list[dict[str, str]]`): The list of prompts to
        narrate columns of tables.
        - **batch_size** (`int`): The specific batch size value to test.

        ## Returns
        - `bool`: The `conversations` fit or not in memory.
        """
        special_indices = self.__get_special_indices(conversations, batch_size)
        adjusted_conversations = [conversations[i] for i in special_indices]

        conv_low_idx = len(adjusted_conversations) // 2 - batch_size // 2
        conv_high_dx = conv_low_idx + batch_size
        output = prompt_pipeline(
            self.pipe,
            adjusted_conversations[conv_low_idx:conv_high_dx],
            batch_size=batch_size,
            context_length=32768,
            max_new_tokens=1,
            temperature=None,
            top_p=None,
            top_k=None,
        )

        torch.cuda.empty_cache()
        gc.collect()

        if output[0][0]["content"] == "":
            del output
            return False
        else:
            del output
            return True

    def __get_special_indices(self, prompts: list[str], batch_size: int) -> list[int]:
        """
        Sorts `prompts` in a specific manner to try to balance the memory load
        for each batch of LLM inferences.

        ## Args
        - **prompts** (`list[str]`): The list of prompts to narrate columns of
        tables.
        - **batch_size** (`int`): The optimal batch size value to be used.

        ## Returns
        - `list[int]`: The "special indices" for the `prompts` given the `batch
        size`.
        """
        # Step 1: Sort the conversations (indices) in decreasing order
        sorted_indices = sorted(
            range(len(prompts)), key=lambda x: len(prompts[x]), reverse=True
        )

        # Step 2: Interleave the indices (longest, shortest, second longest, second shortest, ...)
        final_indices: list[int] = []
        i, j = 0, len(sorted_indices) - 1

        while i <= j:
            if i == j:
                final_indices.append(sorted_indices[i])
                break

            final_indices.append(sorted_indices[i])
            i += 1

            for _ in range(batch_size - 1):
                if i <= j:
                    final_indices.append(sorted_indices[j])
                    j -= 1
                else:
                    break
        return final_indices

    def __block_column_narrations(self, column_narrations: list[str]) -> list[str]:
        """
        Convert column narrations into blocks to try to group multiple narrations
        as much as possible, reducing the amount of embeddings that need to be
        produced.

        ## Args
        - **column_narrations** (`list[str]`): The list of column narrations for
        a set of tables.

        ## Returns
        - `list[str]`: The blocked version of `column_narrations`.
        """
        if isinstance(self.embedding_model, OpenAI):
            tokenizer = tiktoken.encoding_for_model("gpt-4o")
        else:
            tokenizer = self.embedding_model.tokenizer
        merged_column_descriptions = []

        col_idx = 0
        while col_idx < len(column_narrations):
            current_description = column_narrations[col_idx]
            while col_idx + 1 < len(column_narrations):
                combined_description = (
                    current_description + " || " + column_narrations[col_idx + 1]
                )
                if (
                    len(tokenizer.encode(combined_description))
                    < self.EMBEDDING_MAX_TOKENS
                ):
                    current_description = combined_description
                    col_idx += 1
                else:
                    break

            col_idx += 1
            merged_column_descriptions.append(current_description)

        return merged_column_descriptions

    def __generate_row_samples(self, df: pd.DataFrame) -> list[str]:
        """
        Generates row samples for the table `df`. The process is deterministic
        because we set the sampling seed to be the value 0.

        ## Args
        - **df** (`pd.DataFrame`): The specific table to sample rows from.

        ## Returns
        - `list[str]`: The sampled rows
        """
        sample_size = math.ceil(min(len(df), 5))
        selected_df = df.sample(n=sample_size, random_state=0).reset_index(drop=True)

        row_samples = []
        for row_idx, row in selected_df.iterrows():
            formatted_row = " | ".join([f"{col}: {val}" for col, val in row.items()])
            row_samples.append(formatted_row.strip())

        merged_row_samples = self.__block_row_samples(row_samples)
        return merged_row_samples

    def __block_row_samples(self, row_samples: list[str]) -> list[str]:
        """
        Convert row samples into blocks to try to group multiple samples
        as much as possible, reducing the amount of embeddings that need to be
        produced.

        ## Args
        - **row_samples** (`list[str]`): The list of row samples for a set of
        tables.

        ## Returns
        - `list[str]`: The blocked version of `row_samples`.
        """
        if isinstance(self.embedding_model, OpenAI):
            tokenizer = tiktoken.encoding_for_model("gpt-4o")
        else:
            tokenizer = self.embedding_model.tokenizer
        merged_row_samples = []

        row_idx = 0
        while row_idx < len(row_samples):
            current_summary = row_samples[row_idx]
            while row_idx + 1 < len(row_samples):
                combined_summary = current_summary + " || " + row_samples[row_idx + 1]
                if len(tokenizer.encode(combined_summary)) < self.EMBEDDING_MAX_TOKENS:
                    current_summary = combined_summary
                    row_idx += 1
                else:
                    break

            row_idx += 1
            merged_row_samples.append(current_summary)

        return merged_row_samples


if __name__ == "__main__":
    fire.Fire(Summarizer)



"""
prompting_interface.py

This module provides prompting interface for LLMs & OpenAI's embedding models.
"""
import logging

import torch
from openai import OpenAI, OpenAIError
from torch.cuda import OutOfMemoryError
from transformers import set_seed
from transformers.pipelines.text_generation import TextGenerationPipeline
from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from pneuma.utils.logging_config import configure_logging

configure_logging(logger_level=logging.WARNING)
logger = logging.getLogger("PromptingInterface")


def prompt_pipeline(
    pipe: TextGenerationPipeline,
    conversations: list[list[dict[str, str]]],
    batch_size=2,
    context_length=8192,
    max_new_tokens: str = 512,
    do_sample: bool = False,
    top_k: int = 0,
    top_p: float = 1.0,
    penalty_alpha: float = 0.0,
    temperature: float = 0.0,
):
    """
    Prompt the LLM pipeline with a conversation.

    ## Args
    - **pipe** (`TextGenerationPipeline`): An initialized pipeline..
    - **conversations** (`list[list[dict[str, str]]]`): The data type of the model.
    - **batch_size** (`int`): The batch size to process conversations.
    - **context_length** (`int`): The LLM's context length.
    - **max_new_tokens** (`int`): Max tokens to generate.
    - **do_sample** (`bool`): Perform sampling or not.
    - **top_k** (`int`): The number of tokens to consider when sampling.
    - **top_p** (`float`): Minimum cumulative probability of tokens being considered.
    - **penalty_alpha** (`float`): The amount of focus being put to ensure non-repetitiveness.
    - **temperature** (`float`): Control how sharp the distribution (smaller means sharper).

    ## Returns
    - `list[list[dict[str, str]]]`: The conversations with model responses appended.
    """
    generation_configs = {
        "max_new_tokens": max_new_tokens,
        "top_k": top_k,
        "top_p": top_p,
        "do_sample": do_sample,
        "penalty_alpha": penalty_alpha,
        "temperature": temperature,
        "pad_token_id": pipe.tokenizer.eos_token_id,
    }
    __remove_unset_generation_configs(generation_configs)
    try:
        for i in range(len(conversations)):
            conversations[i] = __truncate_conversation_if_necessary(
                pipe.tokenizer, conversations[i], context_length, max_new_tokens
            )
        set_seed(42, deterministic=True)
        answers = pipe(
            conversations, truncation=True, batch_size=batch_size, **generation_configs
        )
        results: list[list[dict[str, str]]] = []
        if isinstance(answers[0], dict):
            answers = [answers]
        for answer in answers:
            results.append(answer[0]["generated_text"])
        return results
    except Exception as error:
        logger.warning(error)
        torch.cuda.empty_cache()
        return [[{"role": "user", "content": ""}]]


def prompt_pipeline_robust(
    pipe: TextGenerationPipeline,
    conversations: list[list[dict[str, str]]],
    batch_size=2,
    context_length=8192,
    max_new_tokens: str = 512,
    do_sample: bool = False,
    top_k: int = 0,
    top_p: float = 1.0,
    penalty_alpha: float = 0.0,
    temperature: float = 0.0,
):
    """
    Prompt the pipeline with a conversation in a robust manner (re-try with lower batch size)

    ## Args
    - **pipe** (`TextGenerationPipeline`): An initialized pipeline.
    - **conversations** (`list[list[dict[str, str]]]`): List of conversations with {role, content}.
    - **batch_size** (`int`): The batch size to process conversations.
    - **context_length** (`int`): The LLM's context length.
    - **max_new_tokens** (`int`): Max tokens to generate.
    - **do_sample** (`bool`): Perform sampling or not.
    - **top_k** (`int`): The number of tokens to consider when sampling.
    - **top_p** (`float`): Minimum cumulative probability of tokens being considered.
    - **penalty_alpha** (`float`): The amount of focus being put to ensure non-repetitiveness.
    - **temperature** (`float`): Control how sharp the distribution (smaller means sharper).

    ## Returns
    - `list[list[dict[str, str]]]`: The conversations with model responses appended.
    """
    generation_configs = {
        "max_new_tokens": max_new_tokens,
        "top_k": top_k,
        "top_p": top_p,
        "do_sample": do_sample,
        "penalty_alpha": penalty_alpha,
        "temperature": temperature,
        "pad_token_id": pipe.tokenizer.eos_token_id,
    }
    __remove_unset_generation_configs(generation_configs)
    batch_size_1_counter = 0
    while True:
        try:
            for i in range(len(conversations)):
                conversations[i] = __truncate_conversation_if_necessary(
                    pipe.tokenizer, conversations[i], context_length, max_new_tokens
                )
            set_seed(42, deterministic=True)
            answers = pipe(
                conversations,
                truncation=True,
                batch_size=batch_size,
                **generation_configs,
            )
            results: list[list[dict[str, str]]] = []
            if isinstance(answers[0], dict):
                answers = [answers]
            for answer in answers:
                results.append(answer[0]["generated_text"])
            return (results, batch_size)
        except OutOfMemoryError:
            batch_size = max(batch_size - 10, 1)
            if batch_size == 1:
                batch_size_1_counter += 1
            if batch_size_1_counter == 5:
                raise OutOfMemoryError(
                    "Your GPU memory is probably limited. Try reducing max_batch_size when initializing Pneuma"
                )
            torch.cuda.empty_cache()
            logger.warning(f"Reducing batch size to {batch_size}")


def prompt_openai_llm(
    llm: OpenAI,
    conversations: list[list[dict[str, str]]],
    model: str = "gpt-4o-mini",
    max_new_tokens: str = 512,
    temperature: float = 0.7,
    top_p: float = 1.0,
    retry_attempts: int = 5,
):
    """
    Replicates the Hugging Face pipeline for text generation using OpenAI.

    ## Args
    - **conversations** (`list[list[dict[str, str]]]`): List of conversations with
    {role, content} format.
    - **max_new_tokens** (`int`): Max tokens to generate.
    - **temperature** (`float`): Control how sharp the distribution (smaller means
    sharper).
    - **top_p** (`float`): Minimum cumulative probability of tokens being considered.
    - **retry_attempts** (`int`): Number of retries in case of failure.

    ### Returns
    - `list[list[dict[str, str]]]`: The conversations with model responses appended.
    """
    for conv in conversations:
        for attempt in range(retry_attempts):
            try:
                response = llm.chat.completions.create(
                    model=model,
                    messages=conv,
                    max_tokens=max_new_tokens,
                    temperature=temperature,
                    top_p=top_p,
                    seed=42,
                )
                conv.append(
                    {
                        "role": "assistant",
                        "content": response.choices[0].message.content,
                    }
                )
                break
            except OpenAIError as e:
                if attempt < retry_attempts - 1:
                    print(
                        f"OpenAI API error: {e}. Retrying ({attempt + 1}/{retry_attempts})..."
                    )
                else:
                    print(f"Failed after {retry_attempts} attempts.")
                    raise e
    return conversations


def prompt_openai_embed(
    embed_model: OpenAI,
    documents: list[str],
    model: str = "text-embedding-3-small",
) -> list[list[float]]:
    """
    Prompts OpenAI's embedding model to encode `documents`.

    ## Args
    - **embed_model** (`OpenAI`): An OpenAI client serving as an embedding model.
    - **documents** (`list[str]`): A list of documents to be encoded.
    - **model** (`str`): The actual embedding model name to use.

    ## Returns
    - `list[list[float]]`: The encoded documents.
    """
    responses = embed_model.embeddings.create(
        input=documents,
        model=model,
    )
    embeddings = [i.embedding for i in responses.data]
    return embeddings


def __remove_unset_generation_configs(generation_configs: dict[str, any]):
    """
    Removes unset generation configs for an LLM pipeline.

    ## Args
    - **generation_configs** (`dict[str, any]`): The generation configs.
    """
    if generation_configs["top_k"] == 0:
        del generation_configs["top_k"]
    if generation_configs["top_p"] == 1.0:
        del generation_configs["top_p"]
    if generation_configs["penalty_alpha"] == 0.0:
        del generation_configs["penalty_alpha"]
    if generation_configs["temperature"] == 0.0:
        del generation_configs["temperature"]


def __truncate_conversation_if_necessary(
    tokenizer: PreTrainedTokenizerBase,
    conversation: list[dict[str, str]],
    context_length: int,
    max_new_tokens: int,
):
    """
    Truncates the text of a conversation according to context_length and
    max_new_tokens.

    ### Args
    - **tokenizer** (`PreTrainedTokenizerBase`): A PreTrainedTokenizerBase from
    HuggingFace.
    - **conversation** (`list[dict[str, str]]`): A conversation for the LLM.
    - **context_length** (`int`): The context length of the LLM.
    - **max_new_tokens** (`int`): Max new tokens to be generated by the LLM.

    ### Returns
    - `str`: The truncated text.
    """
    base = [{"role": "user", "content": ""}]
    base_len = len(
        tokenizer.apply_chat_template(base, tokenize=True, add_generation_prompt=True)
    )
    max_tokens = context_length - base_len - max_new_tokens
    tokens = tokenizer.tokenize(conversation[0]["content"])[:max_tokens]
    conversation[0]["content"] = tokenizer.convert_tokens_to_string(tokens)
    return conversation


if __name__ == "__main__":
    import os

    import torch

    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"

    from pipeline_initializer import initialize_pipeline

    pipe = initialize_pipeline(
        "meta-llama/Meta-Llama-3-8B-Instruct",
        torch.bfloat16,
    )
    conversations = [[{"role": "user", "content": "Tell me about Illinois!"}]]
    output = prompt_pipeline(
        pipe, conversations, 1, 8192, temperature=None, top_p=None, max_new_tokens=20
    )
    print(output)
    assert output[0][-1]["role"] == "assistant"



"""
query_processor.py

This module provides query functionality for users' queries.
"""
import logging
import os

import bm25s
import fire
import Stemmer
from bm25s.tokenization import Tokenized, convert_tokenized_to_string_list
from chromadb_deterministic import PersistentClient
from chromadb_deterministic.api.models.Collection import Collection
from chromadb_deterministic.api.types import QueryResult
from numpy import ndarray
from openai import OpenAI
from scipy.spatial.distance import cosine
from sentence_transformers import SentenceTransformer
from tqdm import tqdm
from transformers import TextGenerationPipeline

from pneuma.utils.logging_config import configure_logging
from pneuma.utils.prompting_interface import (
    prompt_openai_embed,
    prompt_openai_llm,
    prompt_pipeline,
)
from pneuma.utils.response import Response, ResponseStatus

configure_logging()
logger = logging.getLogger("Registrar")


class QueryProcessor:
    """
    Processes queries against generated hybrid indexes.

    This class provides a method to retrieve tables from hybrid indexes,
    helping people find the relevant tables for their tasks.

    ## Attributes
    - **pipe** (`OpenAI | TextGenerationPipeline`): The LLM pipeline for inference.
    - **embedding_model** (`OpenAI | SentenceTransformer`): The model used for
    text embeddings.
    - **stemmer** (`Stemmer`): A stemming tool used for text normalization.
    - **index_path** (`str`): Path to the directory where indexes are stored.
    - **vector_index_path** (`str`): Path for vector-based indexing.
    - **fulltext_index_path** (`str`): Path for full-text search indexing.
    """

    def __init__(
        self,
        llm: OpenAI | TextGenerationPipeline,
        embed_model: OpenAI | SentenceTransformer,
        index_path: str,
    ):
        self.pipe = llm
        self.embedding_model = embed_model
        self.stemmer = Stemmer.Stemmer("english")
        self.vector_index_path = os.path.join(index_path, "vector")
        self.fulltext_index_path = os.path.join(index_path, "fulltext")

    def query(
        self,
        index_name: str,
        queries: str | list[str],
        k: int = 1,
        n: int = 5,
        alpha: float = 0.5,
    ) -> str:
        """
        Retrieves tables for the given `queries` against the index `index_name`.

        ## Args
        - **index_name** (`str`): The name of the index to be retrieved against.
        - **queries** (`str | list[str]`): The query of list of queries to be executed.
        - **k** (`int`): The number of documents associated with the tables to be
        retrieved.
        - **n** (`int`): The multiplicative factor of `k` to pool more relevant
        documents for the hybrid retrieval process.
        - **alpha** (`float`): The weighting factor of the vector and full-text
        retrievers within a hybrid index. Lower `alpha` gives more weight to
        the vector retriever.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        logger.info(f"Querying the index {index_name}")

        if isinstance(queries, str):
            queries = [queries]

        try:
            vector_retriever, fulltext_retriever = self.__get_retrievers(index_name)
        except ValueError:
            error_message = f"Index with name {index_name} does not exist."
            logger.debug(error_message)
            return Response(
                status=ResponseStatus.ERROR,
                message=error_message,
            ).to_json()

        # Retrieve more documents (k * n) to pool more relevant documents
        increased_k = min(k * n, len(fulltext_retriever.corpus))
        queries_tokens: list[Tokenized] = [bm25s.tokenize(query) for query in queries]

        bm25s.tokenize(queries, stemmer=self.stemmer, show_progress=False)

        logger.info("=> Encoding the queries")
        if isinstance(self.embedding_model, OpenAI):
            queries_embeddings = prompt_openai_embed(
                self.embedding_model,
                queries,
            )
        else:
            queries_embeddings: list[list[float]] = self.embedding_model.encode(
                queries, show_progress_bar=False
            ).tolist()

        queries_tables: list[dict[str, str | list[str]]] = []
        for query_idx, query in enumerate(tqdm(queries, desc="Processing the queries")):
            results, scores = fulltext_retriever.retrieve(
                queries[query_idx], k=increased_k, show_progress=False
            )
            bm25_res = (results, scores)
            vec_res = vector_retriever.query(
                query_embeddings=[queries_embeddings[query_idx]], n_results=increased_k
            )

            all_nodes = self.__hybrid_retriever(
                fulltext_retriever,
                vector_retriever,
                bm25_res,
                vec_res,
                increased_k,
                query,
                alpha,
                queries_tokens[query_idx],
                queries_embeddings[query_idx],
            )

            all_nodes = all_nodes[:k]
            tables = []
            for table, score, content in all_nodes:
                table = table.split("_SEP_")[0]
                tables.append(table)
            queries_tables.append(
                {
                    "query": query,
                    "retrieved_tables": tables,
                }
            )

        logger.info("All queries have been processed.")
        return Response(
            status=ResponseStatus.SUCCESS,
            message=f"Queries successful for index {index_name}",
            data=queries_tables,
        ).to_json()

    def __get_retrievers(self, index_name: str) -> tuple[Collection, bm25s.BM25]:
        """
        Get both vector and full-text retrievers of the index `index_name`.

        ## Args
        - **index_name** (`str`): The name of the hybrid index.

        ## Returns
        - `tuple[Collection, bm25s.BM25]`: The vector and full-text retrievers.
        """
        chroma_client = PersistentClient(self.vector_index_path)
        vector_retriever = chroma_client.get_collection(index_name)
        fulltext_retriever = bm25s.BM25.load(
            os.path.join(self.fulltext_index_path, index_name),
            load_corpus=True,
        )
        return (vector_retriever, fulltext_retriever)

    def __hybrid_retriever(
        self,
        bm25_retriever: bm25s.BM25,
        vec_retriever: Collection,
        bm25_res: tuple[ndarray, ndarray],
        vec_res: QueryResult,
        k: int,
        query: str,
        alpha: float,
        query_tokens: Tokenized,
        query_embedding: list[float],
    ) -> list[tuple[str, float, str]]:
        """
        Generates a hybrid index with name `index_name` for a given `table_ids`.

        ## Args
        - **bm25_retriever** (`BM25`): The full-text retriever within the hybrid
        index.
        - **vec_retriever** (`Collection`): The vector retriever within the hybrid
        index.
        - **bm25_res** (`tuple[ndarray, ndarray]`): Retrieval results from the
        full-text retriever.
        - **vec_res** (`QueryResult`): Retrieval results from the vector retriever.
        - **k** (`int`): The number of documents retrieved from both retrievers.
        - **query** (`str`): The query.
        - **alpha** (`float`): The weighting factor of the vector and full-text
        retrievers within a hybrid index. Lower `alpha` gives more weight to
        the vector retriever.
        - **query_tokens** (`Tokenized`): The tokenized query.
        - **query_embeddings** (`list[float]`): The embedding of the query

        ## Returns
        - `list[tuple[str, float, str]]`: The result of the hybrid search.
        """
        vec_ids = {vec_id for vec_id in vec_res["ids"][0]}
        bm25_ids = {node["metadata"]["table"] for node in bm25_res[0][0]}
        dictionary_id_bm25: dict[str, int] = {
            datum["metadata"]["table"]: idx
            for idx, datum in enumerate(bm25_retriever.corpus)
        }

        processed_nodes_bm25 = self.__process_nodes_bm25(
            bm25_res,
            list(vec_ids - bm25_ids),
            dictionary_id_bm25,
            bm25_retriever,
            query_tokens,
        )
        processed_nodes_vec = self.__process_nodes_vec(
            vec_res, list(bm25_ids - vec_ids), vec_retriever, query_embedding
        )

        all_nodes: list[tuple[str, float, str]] = []
        for node_id in sorted(vec_ids | bm25_ids):
            bm25_score_doc = processed_nodes_bm25.get(node_id)
            vec_score_doc = processed_nodes_vec.get(node_id)
            combined_score = alpha * bm25_score_doc[0] + (1 - alpha) * vec_score_doc[0]
            if bm25_score_doc[1] is None:
                doc = vec_score_doc[1]
            else:
                doc = bm25_score_doc[1]

            all_nodes.append((node_id, combined_score, doc))

        sorted_nodes = sorted(all_nodes, key=lambda node: (-node[1], node[0]))[:k]

        reranked_nodes = self.__rerank(sorted_nodes, query)
        return reranked_nodes

    def __process_nodes_bm25(
        self,
        items: tuple[ndarray, ndarray],
        missing_ids: list[str],
        dictionary_id_bm25: dict[str, int],
        bm25_retriever: bm25s.BM25,
        query_tokens: Tokenized,
    ):
        """
        Processes the retrieval results of the full-text retriever for the purpose
        of hybrid search (augment the results with missing IDs of documents
        retrieved from the vector index).

        ## Args
        - **items** (`tuple[ndarray, ndarray]`): Retrieval results from the
        full-text retriever.
        - **missing_ids** (`list[str]`): The IDs available in the retrieval results
        of the vector retriever but not full-text retriever.
        - **dictionary_id_bm25** (`dict[str, int]`): The table-document
        associations within the full-text retriever.
        - **bm25_retriever** (`BM25`): The full-text retriever.
        - **query_tokens** (`Tokenized`): The tokenized query.

        ## Returns
        - `dict[str, tuple[float, str]]`: The processed results representing the
        score and document of each document ID.
        """
        results = [node for node in items[0][0]]
        scores = [node for node in items[1][0]]

        extra_results = [
            bm25_retriever.corpus[dictionary_id_bm25[idx]] for idx in missing_ids
        ]
        extra_scores = [
            bm25_retriever.get_scores(
                convert_tokenized_to_string_list(query_tokens)[0]
            )[dictionary_id_bm25[idx]]
            for idx in missing_ids
        ]

        results.extend(extra_results)
        scores.extend(extra_scores)

        max_score = max(scores)
        min_score = min(scores)

        processed_nodes: dict[str, tuple[float, str]] = {
            node["metadata"]["table"]: (
                (
                    1
                    if min_score == max_score
                    else (scores[i] - min_score) / (max_score - min_score)
                ),
                node["text"],
            )
            for i, node in enumerate(results)
        }
        return processed_nodes

    def __process_nodes_vec(
        self,
        items: QueryResult,
        missing_ids: list[str],
        collection: Collection,
        query_embedding: list[float],
    ):
        """
        Processes the retrieval results of the vector retriever for the purpose
        of hybrid search (augment the results with missing IDs of documents
        retrieved from the full-text index).

        ## Args
        - **items** (`QueryResult`): Retrieval results from the vector retriever.
        - **missing_ids** (`list[str]`): The IDs available in the retrieval results
        of the full-text retriever but not vector retriever.
        - **collection** (`dict[str, int]`): The vector retriever
        - **query_embedding** (`list[float]`): The embedding of the query.

        ## Returns
        - `dict[str, tuple[float, str]]`: The processed results representing the
        score and document of each document ID.
        """
        extra_information = collection.get_fast(
            ids=missing_ids, limit=len(missing_ids), include=["documents", "embeddings"]
        )
        items["ids"][0].extend(extra_information["ids"])
        items["documents"][0].extend(extra_information["documents"])
        items["distances"][0].extend(
            cosine(query_embedding, extra_information["embeddings"][i])
            for i in range(len(missing_ids))
        )

        scores: list[float] = [1 - dist for dist in items["distances"][0]]
        documents: list[str] = items["documents"][0]
        ids: list[str] = items["ids"][0]

        max_score = max(scores)
        min_score = min(scores)

        processed_nodes: dict[str, tuple[float, str]] = {
            ids[idx]: (
                (
                    1
                    if min_score == max_score
                    else (scores[idx] - min_score) / (max_score - min_score)
                ),
                documents[idx],
            )
            for idx in range(len(scores))
        }
        return processed_nodes

    def __rerank(
        self,
        nodes: list[tuple[str, float, str]],
        query: str,
    ) -> list[tuple[str, float, str]]:
        """
        Perform re-ranking of documents against the query. Basically, the `LLM
        Judge` classifies whether a document is relevant or not against the query.

        ## Args
        - **nodes** (`list[tuple[str, float, str]]`): The list of tuples, each of
        which consists of ID, relevance score, and document, resulted from the
        hybrid search mechanism.
        - **query** (`str`): The query.

        ## Returns
        - `dict[str, tuple[float, str]]`: The re-ranked results.
        """
        node_tables = [node[0] for node in nodes]

        relevance_prompts = [
            [
                {
                    "role": "user",
                    "content": self.__get_relevance_prompt(
                        node[2],
                        (
                            "content"
                            if node[0].split("_SEP_")[1].startswith("contents")
                            else "context"
                        ),
                        query,
                    ),
                }
            ]
            for node in nodes
        ]

        if isinstance(self.pipe, OpenAI):
            arguments = prompt_openai_llm(
                self.pipe,
                relevance_prompts,
                max_new_tokens=2,
            )
        else:
            arguments = prompt_pipeline(
                self.pipe,
                relevance_prompts,
                batch_size=2,
                context_length=32768,
                max_new_tokens=2,
                top_p=None,
                temperature=None,
                top_k=None,
            )

        tables_relevance = {
            node_tables[arg_idx]: argument[-1]["content"].lower().startswith("yes")
            for arg_idx, argument in enumerate(arguments)
        }

        new_nodes = [
            (table_name, score, doc)
            for table_name, score, doc in nodes
            if tables_relevance[table_name]
        ] + [
            (table_name, score, doc)
            for table_name, score, doc in nodes
            if not tables_relevance[table_name]
        ]
        return new_nodes

    def __get_relevance_prompt(self, desc: str, desc_type: str, query: str):
        """
        Returns relevance prompts for re-ranking purposes. The prompt format is
        slightly different between content summaries and context (metadata).

        ## Args
        - **desc** (`str`): The description of a table, which is either a content
        summary or context (metadata).
        - **desc_type** (`str`): The description type: content or context.
        - **query** (`str`): The query to be compared against.

        ## Returns
        - `str`: A relevance prompt.
        """
        if desc_type == "content":
            return f"""Given a table with the following columns:
*/
{desc}
*/
and this question:
/*
{query}
*/
Is the table relevant to answer the question? Begin your answer with yes/no."""
        elif desc_type == "context":
            return f"""Given this context describing a table:
*/
{desc}
*/
and this question:
/*
{query}
*/
Is the table relevant to answer the question? Begin your answer with yes/no."""


if __name__ == "__main__":
    fire.Fire(QueryProcessor)



"""
registrar.py

This module handles dataset registration functionality.
"""
import json
import logging
import os

import duckdb
import fire
import pandas as pd

from pneuma.utils.logging_config import configure_logging
from pneuma.utils.response import Response, ResponseStatus
from pneuma.utils.table_status import TableStatus

configure_logging()
logger = logging.getLogger("Registrar")


class Registrar:
    """
    Registers dataset and its context (metadata) to the database.

    This class provides methods to setup the registration system and to
    register tables & context (metadata).

    ## Attributes
    - **db_path** (`str`): Path to the database file for retrieving content
    summaries & context.
    """
    def __init__(self, db_path: str):
        os.makedirs(os.path.dirname(db_path), exist_ok=True)
        self.db_path = db_path

    def setup(self) -> str:
        """
        Setups the database system for registration purposes.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                connection.execute("INSTALL httpfs")
                connection.execute("LOAD httpfs")
                logger.info("HTTPFS installed and loaded")

                connection.sql(
                    """CREATE TABLE IF NOT EXISTS table_status (
                        id VARCHAR PRIMARY KEY,
                        table_name VARCHAR NOT NULL,
                        status VARCHAR NOT NULL,
                        time_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                        creator VARCHAR NOT NULL,
                        hash VARCHAR NOT NULL,
                        )
                    """
                )
                logger.info("Table `table_status` created")

                # Arbitrary auto-incrementing id for contexts and summaries.
                # Change to "CREATE IF NOT EXISTS" on production.
                connection.sql("CREATE SEQUENCE IF NOT EXISTS id_seq START 1")
                logger.info("ID sequence created")

                # DuckDB does not support "ON DELETE CASCADE" so be careful with deletions.
                connection.sql(
                    """CREATE TABLE IF NOT EXISTS table_contexts (
                        id INTEGER DEFAULT nextval('id_seq') PRIMARY KEY,
                        table_id VARCHAR NOT NULL REFERENCES table_status(id),
                        context JSON NOT NULL
                        )
                    """
                )
                logger.info("Table `table_contexts` created")

                # DuckDB does not support "ON DELETE CASCADE" so be careful with deletions.
                connection.sql(
                    """CREATE TABLE IF NOT EXISTS table_summaries (
                        id INTEGER DEFAULT nextval('id_seq') PRIMARY KEY,
                        table_id VARCHAR NOT NULL REFERENCES table_status(id),
                        summary JSON NOT NULL,
                        summary_type VARCHAR NOT NULL,
                        )
                    """
                )
                logger.info("Table `table_summaries` created")

                connection.sql(
                    """CREATE TABLE IF NOT EXISTS indexes (
                        id INTEGER default nextval('id_seq') PRIMARY KEY,
                        name VARCHAR NOT NULL,
                        location VARCHAR NOT NULL,
                        )
                    """
                )
                logger.info("Table `indexes` created")

                connection.sql(
                    """CREATE TABLE IF NOT EXISTS index_table_mappings (
                        index_id INTEGER NOT NULL REFERENCES indexes(id),
                        table_id VARCHAR NOT NULL REFERENCES table_status(id),
                        PRIMARY KEY (index_id, table_id)
                        )
                    """
                )
                logger.info("Table `index_table_mappings` created")

                # TODO: Adjust the response column to the actual response type.
                connection.sql(
                    """CREATE TABLE IF NOT EXISTS query_history (
                        time TIMESTAMP DEFAULT CURRENT_TIMESTAMP PRIMARY KEY,
                        table_id VARCHAR NOT NULL REFERENCES table_status(id), 
                        query VARCHAR NOT NULL,
                        response VARCHAR NOT NULL,
                        querant VARCHAR NOT NULL
                        )
                    """
                )
                logger.info("Table `query_history` created")

                return Response(
                    status=ResponseStatus.SUCCESS,
                    message="Database Initialized.",
                ).to_json()
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error initializing database: {e}",
            ).to_json()

    def add_tables(
        self,
        path: str,
        creator: str,
        source: str = "file",
        s3_region: str = None,
        s3_access_key: str = None,
        s3_secret_access_key: str = None,
        accept_duplicates: bool = False,
    ) -> str:
        """
        Adds tables into the database.

        ## Args
        - **path** (`str`): The path to a specific table file/folder (`CSV` or
        `parquet`).
        - **creator** (`str`): The creator of the file.
        - **source** (`str`): The dataset source (either `file` or `s3`).
        - **s3_region** (`int`): Amazon S3 region.
        - **s3_access_key** (`int`): Amazon S3 access key.
        - **s3_secret_access_key** (`int`): Amazon S3 secret access key.
        - **accept_duplicates** (`bool`): Option to accept duplicate tables or not.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        if source not in ["file", "s3"]:
            return "Invalid source. Please use 'file' or 's3'."

        if source == "s3":
            try:
                with duckdb.connect(self.db_path) as connection:
                    connection.execute(
                        f"""
                    SET s3_region='{s3_region}';
                    SET s3_access_key_id='{s3_access_key}';
                    SET s3_secret_access_key='{s3_secret_access_key}';
                    """
                    )
            except Exception as e:
                return Response(
                    status=ResponseStatus.ERROR,
                    message=f"Error connecting to database: {e}",
                ).to_json()

        if os.path.isfile(path):
            return self.__read_table_file(path, creator, accept_duplicates).to_json()
        if os.path.isdir(path):
            return self.__read_table_folder(path, creator, accept_duplicates).to_json()

        return Response(
            status=ResponseStatus.ERROR,
            message=f"Invalid path: {path}",
        ).to_json()

    def add_metadata(
        self, metadata_path: str, table_id: str = ""
    ) -> str:
        """
        Adds metadata into the database.

        ## Args
        - **metadata_path** (`str`): The path to a specific metadata file/folder
        (`TXT` or `CSV`).
        - **table_id** (`str`): A specific table id associated with the metadata.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        if os.path.isfile(metadata_path):
            return self.__read_metadata_file(
                metadata_path, table_id
            ).to_json()
        if os.path.isdir(metadata_path):
            return self.__read_metadata_folder(
                metadata_path, table_id
            ).to_json()

        return Response(
            status=ResponseStatus.ERROR,
            message=f"Invalid path: {metadata_path}",
        ).to_json()

    def __read_table_file(
        self,
        path: str,
        creator: str,
        accept_duplicates: bool = False,
    ) -> Response:
        """
        Reads a table file (CSV or Parquet), registers it in the database, and
        updates if an existing table has the same ID.

        ## Args
        - **path** (`str`): The path to a specific table file (`CSV` or `parquet`).
        - **creator** (`str`): The creator of the file.
        - **accept_duplicates** (`bool`): Option to allow duplicate tables or not.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                # Index -1 to get the file extension, then slice [1:] to remove the dot.
                file_type = os.path.splitext(path)[-1][1:]

                if file_type not in ["csv", "parquet"]:
                    return Response(
                        status=ResponseStatus.ERROR,
                        message="Invalid file type. Please use 'csv' or 'parquet'.",
                    )

                # If the path contains single quotes, we need to escape them to avoid
                # breaking the SQL query.
                path = path.replace("'", "''")
                if file_type == "csv":
                    name = path.split("/")[-1][:-4]
                    table = connection.sql(
                        f"""SELECT *
                            FROM read_csv(
                                '{path}',
                                auto_detect=True,
                                header=True,
                                ignore_errors=True
                            )"""
                    )
                    table_hash = connection.sql(
                        f"""SELECT md5(string_agg(tbl::text, ''))
                        FROM read_csv(
                            '{path}',
                            auto_detect=True,
                            header=True,
                            ignore_errors=True
                        ) AS tbl"""
                    ).fetchone()[0]
                elif file_type == "parquet":
                    name = path.split("/")[-1][:-8]
                    table = connection.sql(
                        f"""SELECT *
                        FROM read_parquet(
                            '{path}'
                        )"""
                    )
                    table_hash = connection.sql(
                        f"""SELECT md5(string_agg(tbl::text, ''))
                        FROM read_parquet(
                            '{path}'
                        ) AS tbl"""
                    ).fetchone()[0]

                # For ease of keeping track of IDs, we replace backslashes (Windows) with
                # forward slashes (everything else) to make the path (therefore, ID) consistent.
                path = path.replace("\\", "/")

                # We want to avoid double quotes on table names, so we change them to single quotes.
                path = path.replace('"', "''")

                if not accept_duplicates:
                    # Check if table with the same hash already exist
                    table_exist = connection.sql(
                        f"SELECT id FROM table_status WHERE hash = '{table_hash}'"
                    ).fetchone()
                    if table_exist:
                        return Response(
                            status=ResponseStatus.ERROR,
                            message=f"This table already exists in the database with id {table_exist}.",
                        )

                # Check if a table with the same ID already exists
                existing_entry = connection.sql(
                    f"SELECT table_name FROM table_status WHERE id = '{path}'"
                ).fetchone()

                if existing_entry:
                    old_table_name = existing_entry[0]
                    connection.sql(f"DROP TABLE IF EXISTS \"{old_table_name}\"")
                    connection.sql(
                        f"DELETE FROM table_status WHERE id = '{path}'"
                    )

                # The double quote is necessary to consider the path, which may contain
                # full stop that may mess with schema as a single string. Having single quote
                # inside breaks the query, so having the double quote INSIDE the single quote
                # is the only way to make it work.

                # Because we are using double quotes for the table creation, we need to unescape
                # the single quote so the table name fits the ID that is stored in table_status.
                create_path = path.replace("''", "'")
                table.create(f'"{create_path}"')

                connection.sql(
                    f"""INSERT INTO table_status (id, table_name, status, creator, hash)
                    VALUES ('{path}', '{name}', '{TableStatus.REGISTERED}', '{creator}', '{table_hash}')"""
                )
                return Response(
                    status=ResponseStatus.SUCCESS,
                    message=f"Table with ID: {path} has been added to the database.",
                    data={"table_id": path, "table_name": name},
                )
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error connecting to database: {e}",
            )

    def __read_table_folder(
        self, folder_path: str, creator: str, accept_duplicates: bool = False
    ) -> Response:
        """
        Reads a folder and registers all of its tables to the database.

        ## Args
        - **folder_path** (`str`): The path to a folder containing tables.
        - **creator** (`str`): The creator of the file.
        - **accept_duplicates** (`bool`): Option to allow duplicate tables or not.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        logger.info(f"Reading folder {folder_path}")
        paths = [os.path.join(folder_path, f) for f in os.listdir(folder_path)]
        data = []
        for path in paths:
            logger.info(f"=> Processing {path}")

            # If the path is a folder, recursively read the folder.
            if os.path.isdir(path):
                response = self.__read_table_folder(path, creator, accept_duplicates)
                logger.info(response.message)
                data.extend(response.data["tables"])
                continue

            response = self.__read_table_file(path, creator, accept_duplicates)
            logger.info(
                f"==> Processing table {path} {response.status.value}: {response.message}"
            )
            data.append(response.data)

        file_count = len(data)
        return Response(
            status=ResponseStatus.SUCCESS,
            message=f"{file_count} files in folder {folder_path} has been processed.",
            data={"file_count": file_count, "tables": data},
        )

    def __read_metadata_file(
        self, metadata_path: str, table_id: str
    ) -> Response:
        """
        Reads a metadata file (CSV or TXT) and registers it in the database.

        ## Args
        - **metadata_path** (`str`): The path to a specific metadata file (`CSV`
        or `TXT`).
        - **table_id** (`str`): The associated table of the metadata.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        # Index -1 to get the file extension, then slice [1:] to remove the dot.
        file_type = os.path.splitext(metadata_path)[-1][1:]

        if file_type not in ["txt", "csv"]:
            return Response(
                status=ResponseStatus.ERROR,
                message="Invalid file type. Please use 'txt' or 'csv'.",
            )

        if file_type == "txt":
            with open(metadata_path, "r") as f:
                metadata_content = f.read()

            return self.__insert_metadata(metadata_content, table_id)

        if file_type == "csv":
            metadata_df = pd.read_csv(metadata_path)
            data = []
            for index, row in metadata_df.iterrows():
                table_id = row["table_id"]
                metadata_content = row["value"]
                response = self.__insert_metadata(
                    metadata_content, table_id
                )
                logger.info(response.message)
                data.extend(response.data["metadata_ids"])

            return Response(
                status=ResponseStatus.SUCCESS,
                message=f"{len(metadata_df)} metadata entries has been added.",
                data={"file_count": len(data), "metadata_ids": data},
            )

        return None

    def __read_metadata_folder(
        self, metadata_path: str, table_id: str
    ) -> Response:
        """
        Reads a folder and registers all of its metadata to the database.

        ## Args
        - **metadata_path** (`str`): The path to a folder containing metadata files.
        - **table_id** (`str`): The associated table of the metadata files.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        logger.info(f"Reading metadata folder {metadata_path}")
        paths = [os.path.join(metadata_path, f) for f in os.listdir(metadata_path)]
        metadata_ids = []
        for path in paths:
            logger.info(f"=> Processing {path}")

            # If the path is a folder, recursively read the folder.
            if os.path.isdir(path):
                response = self.__read_metadata_folder(path, table_id)
                logger.info(response.message)
                metadata_ids.extend(response.data["metadata_ids"])
                continue

            response = self.__read_metadata_file(path, table_id)
            logger.info(
                f"==> Processing metadata {path} {response.status.value}: {response.message}"
            )
            metadata_ids.extend(response.data["metadata_ids"])

        file_count = len(metadata_ids)
        return Response(
            status=ResponseStatus.SUCCESS,
            message=f"{file_count} files in folder {metadata_path} has been processed.",
            data={"file_count": file_count, "metadata_ids": metadata_ids},
        )

    def __insert_metadata(
        self, metadata_content: str, table_id: str
    ) -> Response:
        """
        Inserts metadata associated with table `table_id` into the database.

        ## Args
        - **metadata_content** (`str`): The content of the metadata.
        - **table_id** (`str`): The associated table of the metadata.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                payload = {
                    "payload": metadata_content.strip(),
                }

                metadata_id = connection.sql(
                    f"""INSERT INTO table_contexts (table_id, context)
                    VALUES ('{table_id}', '{json.dumps(payload)}')
                    RETURNING id"""
                ).fetchone()[0]

                return Response(
                    status=ResponseStatus.SUCCESS,
                    message=f"Metadata created with ID: {metadata_id}",
                    data={"file_count": 1, "metadata_ids": [metadata_id]},
                )
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error connecting to database: {e}",
            )


if __name__ == "__main__":
    fire.Fire(Registrar)



"""
pipeline_initializer.py

This module provides HuggingFace-based LLM pipeline initialization functionality.
"""
from torch import dtype
from transformers import pipeline


def initialize_pipeline(
    model_path: str, torch_dtype: dtype, context_length=8192,
):
    """
    Initializes a HuggingFace text generation pipeline.

    ## Args
    - **model_path** (str): The path of an LLM for the pipeline.
    - **torch_dtype** (dtype): The data type of the LLM.
    - **context_length** (int): The context length of the LLM.

    ## Returns
    - `TextGenerationPipeline`: The pipeline for text generation.
    """
    pipe = pipeline(
        "text-generation",
        model=model_path,
        device_map="auto",
        torch_dtype=torch_dtype,
    )

    pipe.tokenizer.model_max_length = context_length
    return pipe


if __name__ == "__main__":
    import torch
    from transformers.pipelines.text_generation import TextGenerationPipeline

    model_path = "meta-llama/Meta-Llama-3-8B-Instruct"
    pipe = initialize_pipeline(model_path, torch.bfloat16)

    assert type(pipe) == TextGenerationPipeline



from enum import Enum


class SummaryType(Enum):
    """
    Represents the type of a table summary in Pneuma.

    This enum defines the possible summary types.

    ## Members
    - **COLUMN_NARRATION (1)**: LLM narrations for table columns.
    - **ROW_SAMPLE (2)**: Samples of the table rows.
    """
    COLUMN_NARRATION = 1
    ROW_SAMPLE = 2



import logging
import os


def configure_logging(logger_level: str = logging.INFO):
    """
    Configures logging functionality across Pneuma's modules.

    ## Args
    - **logger_level** (int): The minimum level of log messages to be shown.
    """
    LOGGER_LOCATION = os.path.join(os.getcwd(), ".pneuma")
    os.makedirs(LOGGER_LOCATION, exist_ok=True)
    logger = logging.getLogger()
    if logger.hasHandlers():
        logger.handlers.clear()

    logging.basicConfig(
        level=logger_level,
        format="%(asctime)s [%(name)s] [%(levelname)s] %(message)s",
        handlers=[
            logging.FileHandler(os.path.join(LOGGER_LOCATION, "pneuma.log")),
            logging.StreamHandler(),
        ],
    )



"""
pneuma.py

This module serves as an entry point to use `Pneuma`, an LLM-based data discovery
system for tabular data.
"""
import logging
import os
from typing import Optional

import fire
from huggingface_hub import login
from openai import OpenAI
from sentence_transformers import SentenceTransformer
from transformers import TextGenerationPipeline
from torch import bfloat16

from pneuma.index_generator.index_generator import IndexGenerator
from pneuma.query_processor.query_processor import QueryProcessor
from pneuma.registrar.registrar import Registrar
from pneuma.summarizer.summarizer import Summarizer
from pneuma.utils.logging_config import configure_logging
from pneuma.utils.pipeline_initializer import initialize_pipeline

configure_logging()
logger = logging.getLogger("Pneuma")


class Pneuma:
    """
    The entry point of `Pneuma`, combining all modules for ther purpose of LLM-based
    table discovery.

    This class provides end-to-end methods from indexing tables (and their metadata,
    if any) to retrieving tables given users' queries.

    ## Attributes
    - **out_path** (`str`): The output folder of Pneuma.
    - **db_path** (`str`): The database path within the output folder of Pneuma.
    - **index_location** (`str`): The index path within the output folder of Pneuma.
    - **hf_token** (`str`): A HuggingFace User Access Tokens.
    - **openai_api_key** (`str`): An OpenAI API key.
    - **use_local_model** (`bool`): The option to use local or third-party models
    (for now, OpenAI models only as both LLM and embedding model).
    - **llm_path** (`str`): The path or name of a local LLM from HuggingFace.
    - **embed_path** (`str`): The path or name of a local embedding model from
    HuggingFace.
    - **max_llm_batch_size** (`int`): Maximum batch size for the dynamic batch
    size selector to explore.
    - **registrar** (`Registrar`): The dataset regisration module.
    - **summarizer** (`Summarizer`): The dataset summarizer module.
    - **index_generator** (`IndexGenerator`): The index generator module.
    - **query_processor** (`QueryProcessor`): The query processor module.
    - **llm** (`OpenAI | TextGenerationPipeline`): The actual LLM (lazily
    initialized).
    - **embed_model** (`OpenAI | SentenceTransformer`): The actual embedding model
    (lazily initialized).
    """

    def __init__(
        self,
        out_path: Optional[str] = None,
        hf_token: Optional[str] = None,
        openai_api_key: Optional[str] = None,
        use_local_model: bool = True,
        llm_path: str = "Qwen/Qwen2.5-7B-Instruct",
        embed_path: str = "BAAI/bge-base-en-v1.5",
        max_llm_batch_size: int = 50,
    ):
        if out_path is None:
            out_path = os.path.join(os.getcwd(), "pneuma-out")
        os.makedirs(out_path, exist_ok=True)

        self.out_path = os.path.abspath(out_path)
        self.db_path = os.path.join(self.out_path, "storage.db")
        self.index_path = os.path.join(self.out_path, "indexes")

        self.hf_token = hf_token
        self.openai_api_key = openai_api_key
        self.use_local_model = use_local_model
        self.llm_path = llm_path
        self.embed_path = embed_path
        self.max_llm_batch_size = max_llm_batch_size

        self.__hf_login()

        self.registrar: Optional[Registrar] = None  # Handles dataset registration
        self.summarizer: Optional[Summarizer] = None  # Summarizes table contents
        self.index_generator: Optional[IndexGenerator] = (
            None  # Generates document indexes
        )
        self.query_processor: Optional[QueryProcessor] = None  # Handles user queries
        self.llm: Optional[OpenAI | TextGenerationPipeline] = None  # Placeholder for LLM
        self.embed_model: Optional[OpenAI | SentenceTransformer] = None  # Placeholder for embedding model

    def __hf_login(self):
        """Logs into Hugging Face if a token is provided."""
        if self.hf_token:
            try:
                login(self.hf_token)
            except ValueError as e:
                logger.warning(f"HF login failed: {e}")

    def __init_registrar(self):
        """Initializes the Registrar module."""
        self.registrar = Registrar(db_path=self.db_path)

    def __init_summarizer(self):
        """Initializes the Summarizer module."""
        self.__init_llm()
        self.__init_embed_model()
        self.summarizer = Summarizer(
            llm=self.llm,
            embed_model=self.embed_model,
            db_path=self.db_path,
            max_llm_batch_size=self.max_llm_batch_size,
        )

    def __init_index_generator(self):
        """Initializes the IndexGenerator module."""
        self.__init_embed_model()
        self.index_generator = IndexGenerator(
            embed_model=self.embed_model,
            db_path=self.db_path,
            index_path=self.index_path,
        )

    def __init_query_processor(self):
        """Initializes the QueryProcessor module."""
        self.__init_llm()
        self.__init_embed_model()
        self.query_processor = QueryProcessor(
            llm=self.llm,
            embed_model=self.embed_model,
            index_path=self.index_path,
        )

    def __init_llm(self):
        """Initializes the LLM."""
        if self.llm is None:
            if self.use_local_model:
                self.llm = initialize_pipeline(
                    self.llm_path,
                    bfloat16,
                    context_length=32768,
                )
                # Specific setting for batching
                self.llm.tokenizer.pad_token_id = self.llm.model.config.eos_token_id
                self.llm.tokenizer.padding_side = "left"
            else:
                self.llm = OpenAI(api_key=self.openai_api_key)

    def __init_embed_model(self):
        """Initializes the embedding model."""
        if self.embed_model is None:
            if self.use_local_model:
                self.embed_model = SentenceTransformer(self.embed_path)
            else:
                self.embed_model = OpenAI(api_key=self.openai_api_key)

    def setup(self) -> str:
        """Setup Pneuma through its `Registrar` module."""
        if self.registrar is None:
            self.__init_registrar()
        return self.registrar.setup()

    def add_tables(
        self,
        path: str,
        creator: str,
        source: str = "file",
        s3_region: str = None,
        s3_access_key: str = None,
        s3_secret_access_key: str = None,
        accept_duplicates: bool = False,
    ) -> str:
        """
        Registers tables into the database by utilizing the `Registrar` module.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        if self.registrar is None:
            self.__init_registrar()
        return self.registrar.add_tables(
            path,
            creator,
            source,
            s3_region,
            s3_access_key,
            s3_secret_access_key,
            accept_duplicates,
        )

    def add_metadata(
        self,
        metadata_path: str,
        table_id: str = "",
    ) -> str:
        """
        Registers metadata into the database by utilizing the `Registrar` module.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        if self.registrar is None:
            self.__init_registrar()
        return self.registrar.add_metadata(metadata_path, table_id)

    def summarize(self, table_id: str = None) -> str:
        """
        Summarizes the contents of all unsummarized tables or a specific table
        if `table_id` is provided using the `Summarizer` module.

        ## Args
        - **table_id** (`str`): The specific table ID to be summarized.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        if self.summarizer is None:
            self.__init_summarizer()
        return self.summarizer.summarize(table_id)

    def generate_index(
        self, index_name: str, table_ids: list[str] | tuple[str] = None
    ) -> str:
        """
        Generates a hybrid index with name `index_name` for a given `table_ids`
        by utilizing the `IndexGenerator` module.

        ## Args
        - **index_name** (`str`): The name of the index to be generated.
        - **table_ids** (`list[str] | tuple[str]`): The IDs of tables to be indexed
        (to be exact, their content summaries & context/metadata).

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        if self.index_generator is None:
            self.__init_index_generator()
        return self.index_generator.generate_index(index_name, table_ids)

    def query_index(
        self,
        index_name: str,
        queries: list[str] | str,
        k: int = 1,
        n: int = 5,
        alpha: int = 0.5,
    ) -> str:
        """
        Retrieves tables for the given `queries` against the index `index_name`
        by utilizing the `QueryProcessor` module.

        ## Args
        - **index_name** (`str`): The name of the index to be retrieved against.
        - **queries** (`str | list[str]`): The query of list of queries to be executed.
        - **k** (`int`): The number of documents associated with the tables to be
        retrieved.
        - **n** (`int`): The multiplicative factor of `k` to pool more relevant
        documents for the hybrid retrieval process.
        - **alpha** (`float`): The weighting factor of the vector and full-text
        retrievers within a hybrid index. Lower `alpha` gives more weight to
        the vector retriever.

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        if self.query_processor is None:
            self.__init_query_processor()
        return self.query_processor.query(index_name, queries, k, n, alpha)


def main():
    print("Hello from Pneuma's main method!")
    fire.Fire(Pneuma)


if __name__ == "__main__":
    main()



from enum import Enum


class TableStatus(Enum):
    """
    Represents the status of a table in the database of Pneuma.

    This enum defines possible states a table can be in during its lifecycle.

    ## Members
    - **REGISTERED (1)**: The table has been registered but not yet summarized.
    - **SUMMARIZED (2)**: The table has been summarized.
    - **DELETED (3)**: The table has been flagged for removal from the DB.
    """
    REGISTERED = 1
    SUMMARIZED = 2
    DELETED = 3



import json
from enum import Enum
from typing import Any


class ResponseStatus(Enum):
    """
    Represents the status of a response.

    This enum is used to indicate whether an operation was successful or encountered an error.

    ## Members
    - **SUCCESS ("SUCCESS")**: The operation completed successfully.
    - **ERROR ("ERROR")**: The operation failed due to an error.
    """

    SUCCESS = "SUCCESS"
    ERROR = "ERROR"


class Response:
    """
    Represents a standardized API response.

    This class encapsulates a response status, an optional message, and optional data.

    ## Attributes
    - **status** (`ResponseStatus`): The status of the response (either SUCCESS or ERROR).
    - **message** (`str`, optional): An optional message providing additional details.
    - **data** (`Any`, optional): Any additional data to be included in the response.
    """

    def __init__(self, status: ResponseStatus, message: str = None, data: Any = None):
        self.status = status
        self.message = message
        self.data = data

    def to_dict(self) -> dict[str, Any]:
        """Converts a response to a dictionary format."""
        return {"status": self.status.name, "message": self.message, "data": self.data}

    def to_json(self) -> str:
        """Serializes the response as a JSON string."""
        return json.dumps(self.to_dict())



"""
index_generator.py

This module provides indexing functionality for content summaries & context (metadata).
"""

import json
import logging
import os
import time

import bm25s
import duckdb
import fire
import pandas as pd
import Stemmer
import tiktoken
from chromadb_deterministic import Collection, PersistentClient
from chromadb_deterministic.api import ClientAPI
from chromadb_deterministic.db.base import UniqueConstraintError
from openai import OpenAI
from sentence_transformers import SentenceTransformer
from tqdm import tqdm

from pneuma.utils.logging_config import configure_logging
from pneuma.utils.prompting_interface import prompt_openai_embed
from pneuma.utils.response import Response, ResponseStatus
from pneuma.utils.summary_types import SummaryType

configure_logging()
logger = logging.getLogger("IndexGenerator")


class IndexGenerator:
    """
    Generates indexes for content summaries and context (metadata) associated
    with tables.

    This class provides a method to create hybrid---vector & full-text---indexes
    that helps efficiently organize information to be queried later.

    ## Attributes
    - **embedding_model** (`OpenAI | SentenceTransformer`): The model used for
    text embeddings.
    - **db_path** (`str`): Path to the database file for retrieving content
    summaries & context.
    - **index_path** (`str`): Path to the directory where indexes are stored.
    - **stemmer** (`Stemmer`): A stemming tool used for text normalization.
    - **vector_index_path** (`str`): Path for vector-based indexing.
    - **fulltext_index_path** (`str`): Path for full-text search indexing.
    - **EMBEDDING_MAX_TOKENS** (`int`): The maximum number of tokens the embedding
    model supports (hard-coded to 512 for local models and 8191 for OpenAI models).
    """

    def __init__(
        self,
        embed_model: OpenAI | SentenceTransformer,
        db_path: str,
        index_path: str,
    ):
        self.embedding_model = embed_model
        self.db_path = db_path
        self.index_path = index_path
        self.stemmer = Stemmer.Stemmer("english")
        self.vector_index_path = os.path.join(index_path, "vector")
        self.fulltext_index_path = os.path.join(index_path, "fulltext")

        if isinstance(self.embedding_model, OpenAI):
            self.EMBEDDING_MAX_TOKENS = 8191
        else:
            self.EMBEDDING_MAX_TOKENS = 512

    def generate_index(
        self, index_name: str, table_ids: list[str] | tuple[str] = None
    ) -> str:
        """
        Generates a hybrid index with name `index_name` for a given `table_ids`.

        ## Args
        - **index_name** (`str`): The name of the index to be generated.
        - **table_ids** (`list[str] | tuple[str]`): The IDs of tables to be indexed
        (to be exact, their content summaries & context/metadata).

        ## Returns
        - `str`: A JSON string representing the result of the process (`Response`).
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                chroma_client = PersistentClient(self.vector_index_path)
                if table_ids is None:
                    logger.info(
                        "No table ids provided; generating index for all tables"
                    )
                    table_ids = [
                        entry[0]
                        for entry in connection.sql(
                            "SELECT id FROM table_status"
                        ).fetchall()
                    ]

                logger.info(f"Generating index for {len(table_ids)} tables")

                # Step 1a: Generate vector index
                start_time = time.time()
                vector_index_response = self.__generate_vector_index(
                    index_name, chroma_client
                )
                end_time = time.time()
                vector_index_generation_time = end_time - start_time
                if vector_index_response.status == ResponseStatus.ERROR:
                    return vector_index_response.to_json()

                vector_index_id: int = vector_index_response.data["index_id"]
                chroma_collection: Collection = vector_index_response.data["collection"]
                logger.info(vector_index_response.message)

                # Step 1b: Insert documents associated with the tables to the vector index
                vector_insert_response = self.__insert_documents_to_vector_index(
                    vector_index_id, table_ids, chroma_collection
                )
                if vector_insert_response.status == ResponseStatus.ERROR:
                    chroma_client.delete_collection(index_name)
                    return vector_insert_response.to_json()
                logger.info(vector_insert_response.message)

                # Step 2a: Generate full-text index
                start_time = time.time()
                fulltext_index_response = self.__generate_fulltext_index(index_name)
                end_time = time.time()
                fulltext_index_generation_time = end_time - start_time
                if fulltext_index_response.status == ResponseStatus.ERROR:
                    chroma_client.delete_collection(index_name)
                    return fulltext_index_response.to_json()

                fulltext_index_id: int = fulltext_index_response.data["index_id"]
                retriever: bm25s.BM25 = fulltext_index_response.data["retriever"]
                logger.info(fulltext_index_response.message)

                # Step 2b: Insert documents associated with the tables to the FT index
                fulltext_insert_response = self.__insert_documents_to_fulltext_index(
                    fulltext_index_id, table_ids, retriever
                )
                if fulltext_insert_response.status == ResponseStatus.ERROR:
                    chroma_client.delete_collection(index_name)
                    return fulltext_insert_response.to_json()
                logger.info(fulltext_insert_response.message)

                return Response(
                    status=ResponseStatus.SUCCESS,
                    message=f"A hybrid index named {index_name} with id {vector_index_id}"
                    f" (vector index) and {fulltext_index_id} (full-text index)"
                    f" has been created with {len(table_ids)} tables.",
                    data={
                        "table_ids": table_ids,
                        "vector_index_id": vector_index_id,
                        "fulltext_index_id": fulltext_index_id,
                        "vector_index_generation_time": vector_index_generation_time,
                        "fulltext_index_generation_time": fulltext_index_generation_time,
                    },
                ).to_json()
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error connecting to database: {e}",
            ).to_json()

    def __generate_vector_index(
        self, index_name: str, chroma_client: ClientAPI
    ) -> Response:
        """
        Generates a vector index with name `index_name` using `ChromaDB-Deterministic`.

        ## Args
        - **index_name** (`str`): The name of the index to be generated.
        - **chroma_client** (`ClientAPI`): A Client API for `ChromaDB-Deterministic`.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        try:
            chroma_collection = chroma_client.create_collection(
                name=index_name,
                metadata={
                    "hnsw:space": "cosine",
                    "hnsw:random_seed": 42,
                    "hnsw:M": 48,
                },
            )
        except UniqueConstraintError:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Index named {index_name} already exists.",
            )

        try:
            with duckdb.connect(self.db_path) as connection:
                index_id = connection.sql(
                    f"""INSERT INTO indexes (name, location)
                    VALUES ('{index_name}', '{self.vector_index_path}')
                    RETURNING id"""
                ).fetchone()[0]

                return Response(
                    status=ResponseStatus.SUCCESS,
                    message=f"Vector index named {index_name} with id {index_id} has been created.",
                    data={
                        "index_id": index_id,
                        "collection": chroma_collection,
                    },
                )
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error connecting to database: {e}",
            )

    def __insert_documents_to_vector_index(
        self,
        index_id: int,
        table_ids: list[str] | tuple[str],
        chroma_collection: Collection,
    ) -> Response:
        """
        Inserts documents (related to the tables associated with `table_ids`) into
        a vector index.

        ## Args
        - **index_id** (`int`): The database ID of the vector index.
        - **table_ids** (`list[str] | tuple[str]`): The IDs of tables to be indexed
        (to be exact, their content summaries & context/metadata).
        - **chroma_collection** (`Collection`): The vector index.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        documents = []
        ids = []

        for table_id in table_ids:
            logger.info(f"Processing table {table_id}")

            column_narrations = self.__get_content_summaries(
                table_id, SummaryType.COLUMN_NARRATION
            )
            row_samples = self.__get_content_summaries(table_id, SummaryType.ROW_SAMPLE)

            for idx, column_narration in enumerate(column_narrations):
                documents.append(json.loads(column_narration[1])["payload"])
                ids.append(f"{table_id}_SEP_contents_SEP_schema-{idx}")

            for idx, row_sample in enumerate(row_samples):
                documents.append(json.loads(row_sample[1])["payload"])
                ids.append(f"{table_id}_SEP_contents_SEP_row-{idx}")

            contexts = self.__get_table_contexts(table_id)
            if not contexts:
                continue
            contexts = self.__merge_contexts(contexts)
            for context_idx, context in enumerate(contexts):
                documents.append(context)
                ids.append(f"{table_id}_SEP_contexts-{context_idx}")

        if len(documents) == 0:
            return Response(
                status=ResponseStatus.ERROR,
                message="No context and summary entries found for the given table ids.",
            )

        for i in tqdm(range(0, len(documents), 30000)):
            if isinstance(self.embedding_model, OpenAI):
                embeddings = prompt_openai_embed(
                    self.embedding_model,
                    documents[i : i + 30000],
                )
                chroma_collection.add(
                    embeddings=embeddings,
                    documents=documents[i : i + 30000],
                    ids=ids[i : i + 30000],
                )
            else:
                embeddings = self.embedding_model.encode(
                    documents[i : i + 30000],
                    batch_size=10,
                    show_progress_bar=True,
                    device="cuda",
                )
                chroma_collection.add(
                    embeddings=[embed.tolist() for embed in embeddings],
                    documents=documents[i : i + 30000],
                    ids=ids[i : i + 30000],
                )

        insert_df = pd.DataFrame.from_dict(
            {
                "index_id": [index_id] * len(table_ids),
                "table_id": table_ids,
            }
        )

        # So we know which tables are included in this index.
        try:
            with duckdb.connect(self.db_path) as connection:
                connection.sql(
                    """INSERT INTO index_table_mappings (index_id, table_id)
                    SELECT * FROM insert_df""",
                )
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error connecting to database: {e}",
            )

        return Response(
            status=ResponseStatus.SUCCESS,
            message=f"{len(table_ids)} Tables have been inserted to index with id {index_id}.",
        )

    def __generate_fulltext_index(self, index_name: str):
        """
        Generates a full-text index with name `index_name` using `BM25s`.

        ## Args
        - **index_name** (`str`): The name of the index to be generated.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                retriever = bm25s.BM25(corpus=[])
                corpus_tokens = bm25s.tokenize([])
                retriever.index(corpus_tokens, show_progress=False)
                retriever.save(
                    os.path.join(self.fulltext_index_path, index_name), corpus=[]
                )

                index_id = connection.sql(
                    f"""INSERT INTO indexes (name, location)
                    VALUES ('{index_name}', '{self.fulltext_index_path}')
                    RETURNING id"""
                ).fetchone()[0]

                return Response(
                    status=ResponseStatus.SUCCESS,
                    message=f"Fulltext index named {index_name} with id {index_id} has been created.",
                    data={
                        "index_id": index_id,
                        "retriever": retriever,
                    },
                )
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error connecting to database: {e}",
            )

    def __insert_documents_to_fulltext_index(
        self, index_id: int, table_ids: list | tuple, retriever: bm25s.BM25
    ):
        """
        Inserts documents (related to the tables associated with `table_ids`) into
        a full-text index.

        ## Args
        - **index_id** (`int`): The database ID of the full-text index.
        - **table_ids** (`list[str] | tuple[str]`): The IDs of tables to be indexed
        (to be exact, their content summaries & context/metadata).
        - **retriever** (`BM25`): The full-text index.

        ## Returns
        - `Response`: A `Response` object of the process.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                index_name = connection.sql(
                    f"SELECT name FROM indexes WHERE id = {index_id}"
                ).fetchone()[0]

                corpus_json = []
                for table_id in table_ids:
                    logger.info(f"Processing table {table_id}")

                    column_narrations = self.__get_content_summaries(
                        table_id, SummaryType.COLUMN_NARRATION
                    )

                    row_samples = self.__get_content_summaries(
                        table_id, SummaryType.ROW_SAMPLE
                    )

                    for idx, column_narration in enumerate(column_narrations):
                        content = json.loads(column_narration[1])["payload"]
                        corpus_json.append(
                            {
                                "text": content,
                                "metadata": {
                                    "table": f"{table_id}_SEP_contents_SEP_schema-{idx}"
                                },
                            }
                        )

                    for idx, row_sample in enumerate(row_samples):
                        content = json.loads(row_sample[1])["payload"]
                        corpus_json.append(
                            {
                                "text": content,
                                "metadata": {
                                    "table": f"{table_id}_SEP_contents_SEP_row-{idx}"
                                },
                            }
                        )

                    contexts = self.__get_table_contexts(table_id)
                    if not contexts:
                        continue
                    contexts = self.__merge_contexts(contexts)
                    for context_idx, context in enumerate(contexts):
                        corpus_json.append(
                            {
                                "text": context,
                                "metadata": {
                                    "table": f"{table_id}_SEP_contexts-{context_idx}"
                                },
                            }
                        )

                corpus_text = [doc["text"] for doc in corpus_json]
                corpus_tokens = bm25s.tokenize(
                    corpus_text,
                    stopwords="en",
                    stemmer=self.stemmer,
                    show_progress=False,
                )

                retriever.corpus = retriever.corpus + corpus_json
                retriever.index(corpus_tokens, show_progress=False)

                retriever.save(
                    os.path.join(self.fulltext_index_path, index_name),
                    corpus=retriever.corpus,
                )

                insert_df = pd.DataFrame.from_dict(
                    {
                        "index_id": [index_id] * len(table_ids),
                        "table_id": table_ids,
                    }
                )

                # So we know which tables are included in this index.
                connection.sql(
                    """INSERT INTO index_table_mappings (index_id, table_id)
                    SELECT * FROM insert_df""",
                )

                return Response(
                    status=ResponseStatus.SUCCESS,
                    message=f"{len(table_ids)} Tables have been inserted to index with id {index_id}.",
                )
        except Exception as e:
            return Response(
                status=ResponseStatus.ERROR,
                message=f"Error connecting to database: {e}",
            )

    def __get_table_contexts(self, table_id: str) -> list[tuple[str, str]]:
        """
        Retrieves all contexts (metadata) associated with `table_id` from the
        database.

        ## Args
        - **table_id** (`str`): The ID of the table in the database.

        ## Returns
        - `list[tuple[str, str]]`: The contexts and their associated IDs.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                table_id = table_id.replace("'", "''")
                return connection.sql(
                    f"""SELECT id, context FROM table_contexts
                    WHERE table_id='{table_id}'"""
                ).fetchall()
        except Exception as e:
            logger.error(f"Error connecting to database: {e}")

    def __merge_contexts(self, contexts: list[tuple[str, str]]) -> list[str]:
        if isinstance(self.embedding_model, OpenAI):
            tokenizer = tiktoken.encoding_for_model("gpt-4o")
        else:
            tokenizer = self.embedding_model.tokenizer
        table_contexts = [json.loads(context[1])["payload"] for context in contexts]
        processed_contexts = []

        context_idx = 0
        while context_idx < len(table_contexts):
            current_context = table_contexts[context_idx]
            while (context_idx + 1) < len(table_contexts):
                combined_context = (
                    current_context + " | " + table_contexts[context_idx + 1]
                )
                if len(tokenizer.encode(combined_context)) < self.EMBEDDING_MAX_TOKENS:
                    current_context = combined_context
                    context_idx += 1
                else:
                    break

            context_idx += 1
            processed_contexts.append(current_context)

        return processed_contexts

    def __get_content_summaries(
        self, table_id: str, summary_type: SummaryType
    ) -> list[tuple[str, str]]:
        """
        Retrieves all content summaries associated with `table_id` from the
        database.

        ## Args
        - **table_id** (`str`): The ID of the table in the database.
        - **summary_type** (`SummaryType`): The type of summaries to be retrieved
        (either column narration or row sample).

        ## Returns
        - `list[tuple[str, str]]`: The content summaries and their associated IDs.
        """
        try:
            with duckdb.connect(self.db_path) as connection:
                table_id = table_id.replace("'", "''")
                return connection.sql(
                    f"""SELECT id, summary FROM table_summaries
                    WHERE table_id='{table_id}' AND summary_type='{summary_type}'"""
                ).fetchall()
        except Exception as e:
            logger.error(f"Error connecting to database: {e}")


if __name__ == "__main__":
    fire.Fire(IndexGenerator)
